// g++ -Ofast -std=c++11 -lrt -lasound -ljack -lpthread -lboost_iostreams -lboost_system -lboost_filesystem /usr/lib/libportaudio.so joystick/joystick.cc joy2ppm.cpp -o joy2ppm

#include <iostream>
#include <sstream>
#include <fstream>
#include <vector>
#include <tuple>
#include <thread>
#include <mutex>
#include <map>
#include <algorithm>

#include <stdio.h>
#include <math.h>
#include "portaudio.h"
#include "joystick/joystick.hh"
#include "gnuplot-iostream/gnuplot-iostream.h"

using namespace std;

template <typename T> int sgn(T val) {return (T(0) < val) - (val < T(0));}

unsigned sample_rate;
unsigned centerPW;
unsigned deltaPW;
unsigned length;

std::vector<unsigned> ch_width; //us of the up part of each channel
std::mutex ch_mutex;

bool test; //print the joystick movements to help the config
unsigned notch;
std::map<unsigned,int> axis2channel;

static volatile int plot_go;
static volatile int sound_go;
static volatile int joystick_go;
  
static int paCallback( const void *inputBuffer, void *outputBuffer,
                       unsigned long framesPerBuffer,
                       const PaStreamCallbackTimeInfo* timeInfo,
                       PaStreamCallbackFlags statusFlags,
                       void *userData );
void plot();
void sound();
void sample_joystic(Joystick & joystic);

void set_channel_width(size_t i, unsigned us);

int main(int argc, char** argv) {
  ifstream conf_file("joy2ppm.conf");
  string device;
  if (conf_file.is_open()) {
    size_t n_channels=0;
    unsigned start_width = 4000;
    string line;
    while ( getline(conf_file, line) ) {
      stringstream ss(line.substr(0, line.find("#")));
      string item;
      ss >> item;
      if (item == "axis") {
        unsigned axis=0;
        string dummy;
        int channel=0;
        ss >> axis >> dummy >> channel;
        if (channel == 0) continue; //zero is reserved to the start pulse;
        axis2channel.insert(make_pair(axis,channel));
      } else
      if (item == "device")      {ss >> device;} else
      if (item == "sampling")    {ss >> sample_rate;} else
      if (item == "channels")    {ss >> n_channels;} else
      if (item == "pulselength") {ss >> length;} else
      if (item == "centerPW")    {ss >> centerPW;} else
      if (item == "deltaPW")     {ss >> deltaPW;} else
      if (item == "startwidth")  {ss >> start_width;}
    }
    if ((centerPW+deltaPW)*n_channels+start_width > length) {
      cerr << "cannot fit all the channels in the specified length" << endl;
      return 1;
    }
    notch = (centerPW-deltaPW)/2;
    ch_width.assign(n_channels+1, centerPW-notch);
    ch_width[0]=start_width-notch;
  } else {
    cerr << "cannot read the joy2ppm.conf file" << endl;
    return 1;
  }


  bool dojoystick = true;
  // Create an instance of Joystick
  Joystick joystick(device);
  // Ensure that it was found and that we can use it
  if (!joystick.isFound()) {
    cerr << "Joystick \"" << device << "\" not found. Using demo mode.\n";
    dojoystick = false;
  }

  test = false;
  bool doplot = false;
  bool dosound = true;
  for (int i=1; i<argc; i++) {
//    cout << i << ":" << argv[i] << endl;
    if (strcmp(argv[i],"-test")==0)    {test=true;} else
    if (strcmp(argv[i],"-plot")==0)    {doplot=true;} else
    if (strcmp(argv[i],"-nosound")==0) {dosound=false;} 
  }

  ///// now run /////
  std::thread plot_thread;
  std::thread sound_thread;
  std::thread joystick_thread;
  
  if (doplot)     { plot_go     = true; plot_thread     = std::thread(plot); }
  if (dosound)    { sound_go    = true; sound_thread    = std::thread(sound);}
  if (dojoystick) { joystick_go = true; joystick_thread = std::thread(sample_joystic, std::ref(joystick));}
  
  std::this_thread::sleep_for(std::chrono::milliseconds(1000));
  cout << "enter q to quit... ";
  char quit='c';
  while (quit!='q') {
    cin >> quit;
  }

  if (doplot)     { plot_go     = false; plot_thread.join(); }
  if (dosound)    { sound_go    = false; sound_thread.join();}
  if (dojoystick) { joystick_go = false; joystick_thread.join();}
  return 0;
}

////////////////////////////////////////////////////////////////////////////////


unsigned us2sample(long unsigned us) {
  us *= sample_rate;
  us /= 1000000;
  return us;
}

static int paCallback( const void *inputBuffer, void *outputBuffer,
                       unsigned long framesPerBuffer,
                       const PaStreamCallbackTimeInfo* timeInfo,
                       PaStreamCallbackFlags statusFlags,
                       void *userData ) {
  float *out = (float*)outputBuffer;
  (void) inputBuffer; /* Prevent unused variable warning. */
  (void) userData;
  (void) statusFlags;
  (void) timeInfo;

  unsigned cnt = 0;
  std::lock_guard<std::mutex> guard(ch_mutex);
  for (auto w: ch_width) {
    for ( unsigned i = 0; i < us2sample(w); i++) {
      out[cnt++] = 1.0f;
    }
    for ( unsigned i = 0; i < us2sample(notch); i++) {
      out[cnt++] = -1.0f;
    }  
  }
  while (cnt < framesPerBuffer ) {
    out[cnt++] = 1.0f;
  }
  return 0;
}


void sound() {
  PaStream *stream;
  PaError err;

  /* Initialize library before making any other calls. */
  err = Pa_Initialize();
  if( err != paNoError ) goto error;
  
  /* Open an audio I/O stream. */
  err = Pa_OpenDefaultStream( &stream,
                              0,          /* no input channels */
                              1,          /* stereo output */
                              paFloat32,  /* 32 bit floating point output */
                              sample_rate,
                              us2sample(length),        /* frames per buffer */
                              paCallback,
                              nullptr );
  if( err != paNoError ) goto error;

  err = Pa_StartStream( stream );
  if( err != paNoError ) goto error;
  /* sound keeps going using the callback function to refill the buffer */
  while (sound_go)
    Pa_Sleep(100); //ms

  err = Pa_StopStream( stream );
  if( err != paNoError ) goto error;
  err = Pa_CloseStream( stream );
  if( err != paNoError ) goto error;
  Pa_Terminate();
//  printf("Test finished.\n");
  return;
error:
  Pa_Terminate();
  fprintf( stderr, "An error occured while using the portaudio stream\n" );
  fprintf( stderr, "Error number: %d\n", err );
  fprintf( stderr, "Error message: %s\n", Pa_GetErrorText( err ) );
  return;
}

////////////////////////////////////////////////////////////////////////////////

void set_channel(size_t i, unsigned us) {
  std::lock_guard<std::mutex> guard(ch_mutex);
  if ( (i<ch_width.size()) ) {
    ch_width[i]=us;
  }
}

unsigned axis2us(long int a) {
  static long int a_max = 1;
  static long int a_min = -1;
  a_max=std::max(a_max, a);
  a_min=std::min(a_min, a);
  static const long int us_max = centerPW + deltaPW - notch;
  static const long int us_min = centerPW - deltaPW - notch;
  return (us_max-us_min)*(a-a_min)/(a_max-a_min) + us_min;
}

void sample_joystic(Joystick & joystick) {
  while (joystick_go) {
    JoystickEvent event;
    if (joystick.sample(&event)) {
//      if (event.isButton()) {
//        printf("Button %u is %s\n",
//          event.number,
//          event.value == 0 ? "up" : "down");
//      } else
      if (event.isAxis()) {
//        printf("Axis %u is at position %d\n", event.number, event.value);

        auto it = axis2channel.find(event.number);
        
        if (test) {
          if (it != axis2channel.end() ) 
            cout << "moved axis " << it->first << " mapped to channel " 
                 << abs(it->second) << " to val " 
                 << (int) sgn(it->second)*event.value << endl;
          else
            cout << "moved axis " << (unsigned) event.number << " to val "
                 << (int) event.value << endl;
          std::this_thread::sleep_for(std::chrono::milliseconds(1));
        }

        if ( it != axis2channel.end() ) {
          set_channel( abs(it->second), axis2us(sgn(it->second)*event.value));
        }
      }
    }
    std::this_thread::sleep_for(std::chrono::microseconds(100));
  }
}

////////////////////////////////////////////////////////////////////////////////

void plot() {
  Gnuplot gp("gnuplot");
 	gp << "set xrange [0:" << length << "]\nset yrange [-0.5:1.5]\nset key off\nset grid\nset tit 'PPM Signal'\nset xlab 'time [us]'\n";
  while(plot_go) {
    gp << "plot '-' us 1:2 w l lw 2\n";
    std::vector<tuple<unsigned,unsigned>> p;
    p.emplace_back(0,1);
    {
      std::lock_guard<std::mutex> guard(ch_mutex);
      for (auto c : ch_width) {
        p.emplace_back(get<0>(p.back())+c,1);
        p.emplace_back(get<0>(p.back()),0);
        p.emplace_back(get<0>(p.back())+notch,0);
        p.emplace_back(get<0>(p.back()),1);
      }
    }
    p.emplace_back(length,1);
    gp.send1d(p);
    std::this_thread::sleep_for(std::chrono::milliseconds(100));
  }
}
